//##########################################################################
//#                                                                        #
//#        Copyright (C) 2011 by Beijing IRTOUCHSYSTEMS Co., Ltd           #
//#                                                                        #
//#        IRTOUCH Optical USB Touchscreen                                 #    
//#                                                                        #
//#        Create by Yin Xingjie                                             #
//#                                                                        #
//#        Email: yinxj@unitop.cc                                             #
//#                                                                        #
//##########################################################################
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/usb/input.h>
#include <linux/hid.h>
#include <linux/input.h>
#include <linux/cdev.h>
#include <asm/uaccess.h> 
#include <linux/fs.h>

/*
 * Version Information
 */
#define DRIVER_VERSION "v14.2.0.0002 for android"
#define DRIVER_AUTHOR "Beijing IRTOUCH SYSTEMS Co.,Ltd"
#define DRIVER_DESC "IRTOUCH USB touchscreen driver for single point"
#define DRIVER_LICENSE "GPL"
#define TOUCH_DRIVER_NAME "irtouch"

#define MAJOR_DEVICE_NUMBER 47
#define MINOR_DEVICE_NUMBER 191
#define USB_TRANSFER_LENGTH			8
#define TOUCH_DEVICE_NAME "irtouch"

#define CTLCODE_GET_COORDINATE 0xc0
#define CTLCODE_SET_CALIB_PARA_X 0xc1
#define CTLCODE_SET_CALIB_PARA_Y 0xc2
#define CTLCODE_GET_CALIB_PARA_X 0xc3
#define CTLCODE_GET_CALIB_PARA_Y 0xc4
#define CTLCODE_START_CALIB 0xc5
#define CTLCODE_GET_CALIB_STATUS 0xc6
#define CTLCODE_DEVICE_RESET 0xc7
#define CTLCODE_GET_SCR_PARA 0xc8
#define CTLCODE_GET_PRODUCT_SN 0xc9
#define CTLCODE_GET_PRODUCT_ID 0xca
#define CTLCODE_SET_SCAN_MODE 0xcb

#pragma pack(1)
struct irtouch_point
{
	unsigned char	PointId;
	unsigned char	status;
	short	x;
	short	y;
};

struct points
{
	int	x;
	int	y;
};

struct calib_param{
	long	A00;
	long	A01;
	long	A10;
	unsigned char reserve[46];
};

struct	device_config{
	unsigned char deviceID;
	unsigned char monitorID;
	unsigned char reserved[8];
	unsigned char calibrateStatus;
	unsigned char reserved1[47];
};

#pragma pack()

struct irtouchusb {
	char name[128];
	char phys[64];
	struct usb_device *usbdev;
	struct input_dev *dev;
	struct urb *irq;
	unsigned char *data;
	dma_addr_t data_dma;
}; 

struct device_context{
	bool startCalib;
	bool isCalibX;
	bool isCalibY;
	u16  productid;
	struct points points;
	struct calib_param calibX;
	struct calib_param calibY;
	struct device_config devConfig;
};

struct cdev irtouch_cdev;
struct device_context *irtouchdev_context = NULL;
static struct class *irser_class;
static struct usb_driver irtouch_driver;
static struct file_operations irtouch_fops;

static struct usb_class_driver irtouch_class_driver = {
	.name = "irtouch",
	.fops = &irtouch_fops,
	.minor_base = MINOR_DEVICE_NUMBER,
};

static void irtouch_translatePoint(struct irtouch_point pt, int *x, int *y)
{
	*x = ((irtouchdev_context->calibX.A01 * pt.y) / 10000) + ((irtouchdev_context->calibX.A10 * pt.x) / 10000) + irtouchdev_context->calibX.A00;
	*y = ((irtouchdev_context->calibY.A01 * pt.y) / 10000) + ((irtouchdev_context->calibY.A10 * pt.x) / 10000) + irtouchdev_context->calibY.A00;
}

static void irtouch_get_device_param(struct usb_device * udev)
{
	mm_segment_t fs;
	struct file *nef_filp = NULL;
	struct inode *inode = NULL;
	loff_t nef_size = 0;
	u8 buf[256];
	u8 tmp[256];
	loff_t pos = 0;
	ssize_t nread;
	int i = 0, j = 0, k = 0;
	long	cdata[6];
	
	fs = get_fs();
	set_fs(KERNEL_DS);

	nef_filp = filp_open("/sdcard/irtouch/calibParam", O_RDONLY, 0);
	if (IS_ERR(nef_filp)) {
		printk(KERN_ERR "filp_open error: "
		       "Unable to open irtouch calibParam file!\n");
		return;
	}

	if (nef_filp->f_path.dentry) {
		inode = nef_filp->f_path.dentry->d_inode;
	} else {
		printk(KERN_ERR "Can not get valid inode!\n");
		goto out;
	}

	nef_size = i_size_read(inode->i_mapping->host);
	if (nef_size <= 0) {
		printk(KERN_ERR "Invalid irtouch calibParam file size: "
		       "0x%llx\n", nef_size);
		goto out;
	}
	
	nread = vfs_read(nef_filp, buf, 256, &pos);
	if(nread == nef_size) {
		for(i=0; i<nread; i++) {
			if(buf[i] != '\n' || buf[i] != 0x0a)
				tmp[j++] = buf[i];
			else {
				tmp[j] = '\0';
				cdata[k] = simple_strtol(tmp, NULL, 0);
				k++;
				j = 0;
			}
		}
		irtouchdev_context->calibX.A00 = cdata[0];
		printk("calibX.A00=%ld\n", irtouchdev_context->calibX.A00);
		irtouchdev_context->calibX.A01 = cdata[1];
		printk("calibX.A01=%ld\n", irtouchdev_context->calibX.A01);
		irtouchdev_context->calibX.A10 = cdata[2];
		printk("calibX.A10=%ld\n", irtouchdev_context->calibX.A10);
		irtouchdev_context->calibY.A00 = cdata[3];
		printk("calibY.A00=%ld\n", irtouchdev_context->calibY.A00);
		irtouchdev_context->calibY.A01 = cdata[4];
		printk("calibY.A01=%ld\n", irtouchdev_context->calibY.A01);
		irtouchdev_context->calibY.A10 = cdata[5];
		printk("calibY.A10=%ld\n", irtouchdev_context->calibY.A10);
		irtouchdev_context->devConfig.calibrateStatus = 1;
	}
	
out:
	filp_close(nef_filp, current->files);
	set_fs(fs);
}

static int irtouch_open(struct inode * inode, struct file * filp)
{
	struct irtouchusb * irtouch;
	struct usb_interface * intf;
	
	dbg("IRTOUCH:  init_open");

	intf = usb_find_interface(&irtouch_driver, MINOR_DEVICE_NUMBER);
	if(!intf)
	{
		err("IRTOUCH:    %s usb_find_interface ERROR!",__func__);
		return -1;
	}

	irtouch = usb_get_intfdata(intf);
	if(!irtouch)
	{
		err("IRTOUCH:    %s touch is NULL!",__func__);
		return -1;
	}

	filp->private_data = irtouch;

	return 0;
}

static int irtouch_release(struct inode * inode, struct file * filp)
{
	dbg("IRTOUCH:  irtouch_release");
	return 0;
}

static long irtouch_ioctl(struct file * filp, unsigned int ctl_code, unsigned long ctl_param)
{
	unsigned char value = 0;
	int ret = -1;
	struct irtouchusb *irtouch;

	dbg("IRTOUCH:    irtouch_ioctl");

	if(NULL == filp->private_data)
	{
		err("IRTOUCH:    %s private_data is NULL!",__func__);
		return ret;
	}

	irtouch = filp->private_data;

	switch(ctl_code)
	{
		case CTLCODE_START_CALIB:				
			ret = copy_from_user(&value, (unsigned char *)ctl_param, sizeof(unsigned char));
			if(ret == 0)
			{
				if(value == 0x01)
				{
					irtouchdev_context->startCalib = true;
				}
				else
				{
					irtouchdev_context->startCalib = false;
				}
			}	
			
			break;

		case CTLCODE_GET_COORDINATE:
			ret = copy_to_user((struct points *)ctl_param, &irtouchdev_context->points, sizeof(struct points));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_COORDINATE>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_GET_CALIB_STATUS:
			{
				int status = irtouchdev_context->devConfig.calibrateStatus;
				ret = copy_to_user((int *)ctl_param, &status, sizeof(int));
				if(ret != 0)
				{
					err("IRTOUCH:    %s <CTLCODE_GET_CALIB_STATUS>copy_to_user failed!",__func__);
				}
				break;
			}
	
		case CTLCODE_SET_CALIB_PARA_X:
			ret = copy_from_user(&irtouchdev_context->calibX, (struct calib_param *)ctl_param, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_X>copy_to_user failed!",__func__);
			}
			else
			{
				irtouchdev_context->isCalibX = true;
			}
			break;

		case CTLCODE_SET_CALIB_PARA_Y:
			ret = copy_from_user(&irtouchdev_context->calibY, (struct calib_param *)ctl_param, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_SET_CALIB_PARA_Y>copy_to_user failed!",__func__);
			}
			else
			{
				irtouchdev_context->isCalibY = true;
			}
			break;

		case CTLCODE_GET_CALIB_PARA_X:
			ret = copy_to_user((struct calib_param *)ctl_param, &irtouchdev_context->calibX, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_CALIB_PARA>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_GET_CALIB_PARA_Y:
			ret = copy_to_user((struct calib_param *)ctl_param, &irtouchdev_context->calibY, sizeof(struct calib_param));
			if(ret != 0)
			{
				err("IRTOUCH:    %s <CTLCODE_GET_CALIB_PARA>copy_to_user failed!",__func__);
			}
			break;

		case CTLCODE_GET_PRODUCT_ID:
			{
				int status = irtouchdev_context->productid;
				ret = copy_to_user((int *)ctl_param, &status, sizeof(int));
				if(ret != 0)
				{
					err("IRTOUCH:    %s <CTLCODE_GET_PRODUCT_ID>copy_to_user failed!",__func__);
				}
				break;
			}
		
		case CTLCODE_DEVICE_RESET:
			ret = copy_from_user(&value, (unsigned char *)ctl_param, sizeof(unsigned char));
			if(ret == 0)
			{
				if(value == 0x01)
				{
					irtouchdev_context->devConfig.calibrateStatus = 0;
				}
			}
			break;
		default:
			break;
	}

	if(irtouchdev_context->isCalibX && irtouchdev_context->isCalibY)
	{
		irtouchdev_context->isCalibX = false;
		irtouchdev_context->isCalibY = false;
		irtouchdev_context->devConfig.calibrateStatus = 1;
	}
	return 0;
}

static void processInputData(struct irtouchusb *usbtouch,
                                 unsigned char *pkt, int len)
{
	int x = 0, y = 0, touch = 0;
	struct irtouch_point raw_point;
	
	x = ((pkt[3] << 8) | pkt[2]) << 3;
	y = ((pkt[5] << 8) | pkt[4]) << 3;
	touch = (pkt[1] & 0x03) ? 1 : 0;

	if(touch)
	{
		if(irtouchdev_context->startCalib)
		{
			irtouchdev_context->points.x = x;
			irtouchdev_context->points.y = y;
		}
	}
	
	if(irtouchdev_context->devConfig.calibrateStatus && !irtouchdev_context->startCalib)
	{
		raw_point.x = x;
		raw_point.y = y;
		irtouch_translatePoint(raw_point,&x,&y);
	}

	printk("status=%d, x=%d, y=%d!\n", touch, x, y);
	input_report_key(usbtouch->dev, BTN_TOUCH, touch);
	input_report_abs(usbtouch->dev, ABS_X, x);
	input_report_abs(usbtouch->dev, ABS_Y, y);
	input_sync(usbtouch->dev);
}

static void irtouch_irq(struct urb *urb)
{
	struct irtouchusb *ir = urb->context;
	int status;

	switch (urb->status) {
		case 0:            /* success */
			break;
		case -ECONNRESET:  /* unlink */
		case -ENOENT:
		case -ESHUTDOWN:
			return;    /* -EPIPE:  should clear the halt */
		default:	   /* error */
			err("%s - nonzero urb status received: %d",
		    __func__, urb->status);
			goto resubmit;
	}
	
	processInputData(ir, ir->data, USB_TRANSFER_LENGTH);

resubmit:
	status = usb_submit_urb (urb, GFP_ATOMIC);
	if (status)
		err ("can't resubmit intr, %s-%s/input0, status %d",
				ir->usbdev->bus->bus_name,
				ir->usbdev->devpath, status);
}

static struct file_operations irtouch_fops = {
	.owner = THIS_MODULE,
	.open = irtouch_open,
	.unlocked_ioctl = irtouch_ioctl,
	.release = irtouch_release,
};

static int irtouch_open_device(struct input_dev *dev)
{
	struct irtouchusb *ir = input_get_drvdata(dev);

	ir->irq->dev = ir->usbdev;
	if (usb_submit_urb(ir->irq, GFP_KERNEL))
		return -EIO;

	return 0;
}

static void irtouch_close_device(struct input_dev *dev)
{
	struct irtouchusb *ir = input_get_drvdata(dev);

	usb_kill_urb(ir->irq);
}

static bool irtouch_mkdev(void)
{
	int retval;

	//create device node
	dev_t devno = MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER);

	retval = register_chrdev_region(devno,1,TOUCH_DEVICE_NAME);
	if(retval < 0)
	{
		err("IRTOUCH:    %s register chrdev error.",__func__);
		return false;
	}

	cdev_init(&irtouch_cdev,&irtouch_fops);
	irtouch_cdev.owner = THIS_MODULE;
	irtouch_cdev.ops = &irtouch_fops;
	retval = cdev_add(&irtouch_cdev,devno,1);

	if(retval)
	{
		err("IRTOUCH:    %s  Adding char_reg_setup_cdev error=%d",__func__,retval);
		return false;
	}

	irser_class = class_create(THIS_MODULE, TOUCH_DEVICE_NAME);
	if(IS_ERR(irser_class))
	{
		err("IRTOUCH:    %s class create failed.",__func__);
		return false;
	}
	
	device_create(irser_class,NULL,MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER),NULL,TOUCH_DEVICE_NAME);

	printk("IRTOUCH:    %s success!\n",__func__);
	return true;
}

static int irtouch_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct usb_host_interface *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct irtouchusb *ir;
	struct input_dev *input_dev;
	int pipe;
	int error = -ENOMEM;
	
	interface = intf->cur_altsetting;

	if (interface->desc.bNumEndpoints != 1)
		return -ENODEV;
	
	endpoint = &interface->endpoint[0].desc;
	if (!(endpoint->bEndpointAddress & USB_DIR_IN))
		return -ENODEV;
	if ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) != USB_ENDPOINT_XFER_INT)
		return -ENODEV;
	
	pipe = usb_rcvintpipe(dev, endpoint->bEndpointAddress);
	
	if (!(irtouchdev_context = kmalloc(sizeof(struct device_context), GFP_KERNEL)))  
		return -ENOMEM; 
	memset(irtouchdev_context,0,sizeof(struct device_context));
	
	ir = kzalloc(sizeof(struct irtouchusb), GFP_KERNEL);
	input_dev = input_allocate_device();
	if (!ir || !input_dev)
		goto fail1;

	ir->data = usb_alloc_coherent(dev, USB_TRANSFER_LENGTH, GFP_ATOMIC, &ir->data_dma);
	if (!ir->data)
		goto fail1;
	
	ir->irq = usb_alloc_urb(0, GFP_KERNEL);
	if (!ir->irq)
		goto fail2;
	
	ir->usbdev = dev;
	ir->dev = input_dev;

	if (dev->manufacturer)
		strlcpy(ir->name, dev->manufacturer, sizeof(ir->name));

	if (dev->product) {
		if (dev->manufacturer)
			strlcat(ir->name, " ", sizeof(ir->name));
		strlcat(ir->name, dev->product, sizeof(ir->name));
	}

	if (!strlen(ir->name))
		snprintf(ir->name, sizeof(ir->name),
				"IRTOUCH single-touch touchscreen %04x:%04x",
				le16_to_cpu(dev->descriptor.idVendor),
				le16_to_cpu(dev->descriptor.idProduct));

	usb_make_path(dev, ir->phys, sizeof(ir->phys));
	strlcat(ir->phys, "/input0", sizeof(ir->phys));

	input_dev->name = ir->name;
	input_dev->phys = ir->phys;
	usb_to_input_id(dev, &input_dev->id);
	input_dev->dev.parent = &intf->dev;

	input_dev->evbit[0] = BIT_MASK(EV_KEY) | BIT_MASK(EV_ABS);
	input_dev->keybit[BIT_WORD(BTN_TOUCH)] = BIT_MASK(BTN_TOUCH);
	input_set_abs_params(input_dev, ABS_X, 0, 32767, 0, 0);
	input_set_abs_params(input_dev, ABS_Y, 0, 32767, 0, 0);
	//input_set_abs_params(input_dev, ABS_PRESSURE, 0,255, 0, 0);
	
	input_set_drvdata(input_dev, ir);

	input_dev->open = irtouch_open_device;
	input_dev->close = irtouch_close_device;
	
	usb_fill_int_urb(ir->irq, dev, pipe, ir->data, USB_TRANSFER_LENGTH,
			irtouch_irq, ir, endpoint->bInterval);
	ir->irq->transfer_dma = ir->data_dma;
	ir->irq->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

	error = input_register_device(ir->dev);

	usb_set_intfdata(intf, ir);

	error = usb_register_dev(intf, &irtouch_class_driver);
	if(error)
	{
		err("IRTOUCH:     %s Not able to get a minor for this device.",__func__);
		usb_set_intfdata(intf,NULL);
		goto fail3;
	}

	msleep(1000);
	
	irtouch_get_device_param(dev);
	irtouch_mkdev();
	irtouchdev_context->productid = dev->descriptor.idProduct;
	printk("IRTOUCH id=0x%x: %s success!\n", irtouchdev_context->productid, __func__);
	return 0;

fail3:
	usb_free_urb(ir->irq);
fail2:
	usb_free_coherent(dev, USB_TRANSFER_LENGTH, ir->data, ir->data_dma);
	
fail1:
	input_free_device(input_dev);
	kfree(ir);
	kfree(irtouchdev_context);
	return -ENOMEM;
}

static void irtouch_disconnect(struct usb_interface *intf)
{
	struct irtouchusb *ir = usb_get_intfdata (intf);

	dev_t devno = MKDEV(MAJOR_DEVICE_NUMBER,MINOR_DEVICE_NUMBER);

	cdev_del(&irtouch_cdev);
	unregister_chrdev_region(devno,1);

	device_destroy(irser_class,devno);
	class_destroy(irser_class);

	usb_deregister_dev(intf, &irtouch_class_driver);

	usb_set_intfdata(intf, NULL);
	if(ir) 
	{
		usb_kill_urb(ir->irq);
		input_unregister_device(ir->dev);
		usb_free_urb(ir->irq);
		usb_free_coherent(interface_to_usbdev(intf), USB_TRANSFER_LENGTH, ir->data, ir->data_dma);
		kfree(ir);
	}

	if(irtouchdev_context)
	{
		kfree(irtouchdev_context);
	}
}

static struct usb_device_id irtouch_id_table [] = {
	{ USB_DEVICE(0x6615, 0x0001) }, 
	{ USB_DEVICE(0x595a, 0x0001) }, 
	{ }    /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, irtouch_id_table);

static struct usb_driver irtouch_driver = {
	.name        = "IRTOUCH",
	.probe        = irtouch_probe,
	.disconnect    = irtouch_disconnect,
	.id_table    = irtouch_id_table,
};

static int __init irtouch_init(void)
{
	int retval = usb_register(&irtouch_driver);
	if (retval == 0)
		printk(KERN_INFO KBUILD_MODNAME ": " DRIVER_VERSION ":" DRIVER_DESC);
	return retval;
}

static void __exit irtouch_exit(void)
{
	usb_deregister(&irtouch_driver);
}

module_init(irtouch_init);
module_exit(irtouch_exit);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE(DRIVER_LICENSE);
MODULE_VERSION(DRIVER_VERSION);
